import abc
import datetime
import time
import typing

import praw

import yfinance

from . import model


class DataProxy:
    def __init__(self, symbol: str, *args, **kwargs) -> None:
        pass

    @abc.abstractclassmethod
    def get_info(self) -> model.StockInformation:
        ...

    @abc.abstractmethod
    def get_last_price(self) -> float:
        ...

    @abc.abstractmethod
    def get_stock_price_history(
        self, interval: datetime.timedelta, period: datetime.timedelta
    ) -> typing.List[model.StockHistoryDatum]:
        ...

    @abc.abstractmethod
    def get_option_chain(self, date: datetime.datetime) -> model.OptionChain:
        ...

    @abc.abstractmethod
    def get_full_option_chain(
        self
    ) -> typing.Dict[datetime.datetime, model.OptionChain]:
        ...

    @abc.abstractmethod
    def get_reddit_threads(
        self, subreddits: typing.List, time_filter: str = "day", sort: str = "hot"
    ) -> typing.List[praw.models.Submission]:
        ...


class RedditFetcher(DataProxy):
    reddit_client: praw.Reddit

    def __init__(
        self, symbol: str, *args, reddit_client: praw.Reddit = None, **kwargs
    ) -> None:
        self.symbol = symbol
        self.reddit_client = reddit_client

    def get_reddit_threads(
        self, subreddits: typing.List, time_filter: str = "day", sort: str = "hot"
    ) -> typing.List[praw.models.Submission]:
        return [
            p
            for subreddit in subreddits
            for p in self.reddit_client.subreddit(subreddit).search(
                self.symbol, time_filter=time_filter, sort=sort
            )
        ]


class YFinanceProvider(DataProxy):

    ticker: typing.Optional[yfinance.Ticker]
    proxy: typing.Optional[str]

    def __init__(self, symbol: str, *args, proxy: str = None, **kwargs):
        self.symbol = symbol
        self.ticker = None
        self.proxy = proxy

    def get_info(self) -> model.StockInformation:
        stock_information = model.StockInformation(
            time=int(time.time()), symbol=self.symbol
        )
        ticker_info = self.get_ticker().info
        stock_information.average_volume_10_days = ticker_info.get(
            "averageVolume10days", None
        )
        stock_information.average_volume = ticker_info.get("averageVolume", None)
        stock_information.short_ratio = ticker_info.get("shortRatio", None)
        stock_information.short_percent_float = ticker_info.get(
            "shortPercentOfFloat", None
        )
        stock_information.shares_short = ticker_info.get("sharesShort", None)
        stock_information.shares_short_previous_month_date = ticker_info.get(
            "sharesShortPreviousMonthDate", None
        )
        stock_information.shares_short_prior_month = ticker_info.get(
            "sharesShortPriorMonth", None
        )
        stock_information.shares_percent_shares_out = ticker_info.get(
            "sharesPercentSharesOut", None
        )
        stock_information.shares_outstanding = ticker_info.get(
            "sharesOutStanding", None
        )
        stock_information.logo_url = ticker_info.get("logo_url", None)
        stock_information.market_cap = ticker_info.get("marketCap", None)
        stock_information.shares_held_percent_insiders = ticker_info.get(
            "heldPercentInsiders", None
        )
        stock_information.shares_held_percent_institution = ticker_info.get(
            "heldPercentInstitutions", None
        )
        stock_information.long_name = ticker_info.get("longName", None)
        stock_information.shares_implied_outstanding = ticker_info.get(
            "impliedSharesOutstanding", None
        )
        stock_information.shares_outstanding = ticker_info.get(
            "sharesOutstanding", None
        )
        stock_information.float_shares = ticker_info.get("floatShares", None)
        return stock_information

    def get_ticker(self) -> yfinance.Ticker:
        if not self.ticker:
            self.ticker = yfinance.Ticker(self.symbol)
        return self.ticker

    def timedelta_period_to_str(self, period: datetime.timedelta) -> str:

        if period.days == 1:
            return "1d"
        elif period.days == 5:
            return "5d"
        elif period.days == 30:
            return "1mo"
        elif period.days == 30 * 3:
            return "3mo"
        elif period.days == 30 * 6:
            return "6mo"
        elif period.days == 365:
            return "1y"
        elif period.days == 2 * 365:
            return "2y"
        elif period.days == 5 * 365:
            return "5y"
        return ""

    def timedelta_interval_to_str(self, interval: datetime.timedelta) -> str:

        if interval.seconds == 60:
            return "1m"
        elif interval.seconds == 60 * 2:
            return "2m"
        elif interval.seconds == 5 * 60:
            return "5m"
        elif interval.seconds == 15 * 60:
            return "15m"
        elif interval.seconds == 30 * 60:
            return "30m"
        elif interval.seconds == 60 * 60:
            return "60m"
        elif interval.seconds == 90 * 60:
            return "90m"
        elif interval.days == 1:
            return "1d"
        elif interval.days == 5:
            return "5d"
        elif interval.days == 7:
            return "1wk"
        elif interval.days == 30:
            return "1mo"
        elif interval.days == 90:
            return "3mo"
        return ""

    def get_last_price(self) -> float:
        data = self.get_ticker().history()
        return data.tail(1)["Close"].iloc[0]

    def get_stock_price_history(
        self, interval: datetime.timedelta, period: datetime.timedelta
    ) -> typing.List[model.StockHistoryDatum]:

        history = self.get_ticker().history(
            period=self.timedelta_period_to_str(period),
            interval=self.timedelta_interval_to_str(interval),
            proxy=self.proxy,
        )

        earnings = self.get_ticker().get_earnings(proxy=self.proxy, as_dict=True)

        currency = earnings["financialCurrency"]

        stock_history_data = []
        for index, row in history.iterrows():
            utc_data = index.replace(tzinfo=datetime.timezone.utc)

            datum = model.StockHistoryDatum(
                symbol=self.symbol,
                currency=currency,
                time=utc_data,
                open_value=row.Open,
                close_value=row.Close,
                high=row.High,
                low=row.Low,
                volume=row.Volume,
                dividends=row.Dividends,
                stock_splits=row["Stock Splits"],
            )
            stock_history_data.append(datum)
        return stock_history_data

    def get_option_chain(
        self, date: typing.Optional[datetime.datetime] = None
    ) -> model.OptionChain:
        if date:
            date = date.replace(hour=23, minute=59, second=59)
            option_chain = self.get_ticker().option_chain(
                date.strftime("%Y-%m-%d"), proxy=self.proxy
            )
            current_stock_price = self.get_last_price()
        else:
            option_chains_date = self.get_ticker().options

            option_chain = self.get_ticker().option_chain(
                option_chains_date[0], proxy=self.proxy
            )
            current_stock_price = self.get_last_price()
            date = datetime.datetime.strptime(option_chains_date[0], "%Y-%m-%d")
            date = date.replace(hour=23, minute=59, second=59)
        calls = [
            model.OptionChainDatum(
                type=model.OptionType.CALL,
                strike=row.strike,
                bid=row.bid,
                current_stock_price=current_stock_price,
                option_date=date,
                last_trade_date=row.lastTradeDate,
                last_price=row.lastPrice,
                open_interest=row.openInterest,
                implied_volatility=row.impliedVolatility,
                currency=row.currency,
            )
            for idx, row in option_chain.calls.iterrows()
        ]
        puts = [
            model.OptionChainDatum(
                type=model.OptionType.PUT,
                strike=row.strike,
                bid=row.bid,
                current_stock_price=current_stock_price,
                option_date=date,
                last_trade_date=row.lastTradeDate,
                last_price=row.lastPrice,
                open_interest=row.openInterest,
                implied_volatility=row.impliedVolatility,
                currency=row.currency,
            )
            for idx, row in option_chain.puts.iterrows()
        ]

        return model.OptionChain(calls=calls, puts=puts)

    def get_full_option_chain(
        self
    ) -> typing.Dict[datetime.datetime, model.OptionChain]:
        ret = {}
        current_stock_price = self.get_last_price()
        for option_expire in self.get_ticker().options:
            date = datetime.datetime.strptime(option_expire, "%Y-%m-%d")
            date = date.replace(hour=23, minute=59, second=59)
            option_chain = self.get_ticker().option_chain(
                option_expire, proxy=self.proxy
            )
            calls = [
                model.OptionChainDatum(
                    type=model.OptionType.CALL,
                    strike=row.strike,
                    bid=row.bid,
                    current_stock_price=current_stock_price,
                    option_date=date,
                    last_trade_date=row.lastTradeDate,
                    last_price=row.lastPrice,
                    open_interest=row.openInterest,
                    implied_volatility=row.impliedVolatility,
                    currency=row.currency,
                )
                for idx, row in option_chain.calls.iterrows()
            ]
            puts = [
                model.OptionChainDatum(
                    type=model.OptionType.PUT,
                    strike=row.strike,
                    bid=row.bid,
                    current_stock_price=current_stock_price,
                    option_date=date,
                    last_trade_date=row.lastTradeDate,
                    last_price=row.lastPrice,
                    open_interest=row.openInterest,
                    implied_volatility=row.impliedVolatility,
                    currency=row.currency,
                )
                for idx, row in option_chain.puts.iterrows()
            ]

            ret[date] = model.OptionChain(calls=calls, puts=puts)
        return ret


def combine_providers(classes: typing.List[DataProxy]) -> DataProxy:
    class CombinedProvider(*classes):
        def __init__(self, *args, **kwargs) -> None:
            for c in classes:
                c.__init__(self, *args, **kwargs)

    return CombinedProvider



import dataclasses
import datetime
import enum
import math
import typing

import numpy as np
import scipy.stats as scipy_stats

import py_vollib.ref_python.black.greeks.analytical as vollib_a

from . import exceptions


class OptionType(enum.Enum):
    UNDEFINED = 0
    CALL = 1
    PUT = 2


class CorrelationHistoryDatum(typing.NamedTuple):
    time: datetime.datetime
    value: float


class StockHistoryDatum(typing.NamedTuple):
    time: datetime.datetime
    symbol: str
    currency: str
    open_value: float
    close_value: float
    high: float
    low: float
    volume: float
    dividends: float
    stock_splits: float


@dataclasses.dataclass
class StockInformation:
    time: int
    symbol: typing.Optional[str] = None
    logo_url: typing.Optional[str] = None
    long_name: typing.Optional[str] = None
    average_volume_10_days: typing.Optional[float] = None
    average_volume: typing.Optional[float] = None
    market_cap: typing.Optional[float] = None
    float_shares: typing.Optional[float] = None
    short_ratio: typing.Optional[float] = None
    short_percent_float: typing.Optional[float] = None
    shares_short: typing.Optional[float] = None
    shares_short_previous_month_date: typing.Optional[float] = None
    shares_short_prior_month: typing.Optional[float] = None
    shares_percent_shares_out: typing.Optional[float] = None
    shares_outstanding: typing.Optional[float] = None
    shares_implied_outstanding: typing.Optional[float] = None
    shares_held_percent_insiders: typing.Optional[float] = None
    shares_held_percent_institution: typing.Optional[float] = None


class OptionChainDatum(typing.NamedTuple):
    type: OptionType
    strike: float
    current_stock_price: float
    bid: float
    open_interest: float
    currency: str
    last_price: float
    implied_volatility: float
    option_date: datetime.datetime
    last_trade_date: datetime.datetime

    def delta(self, risk: float = 0.0) -> float:
        if self.type == OptionType.UNDEFINED:
            raise exceptions.OptionTypeIsUndefined()
        expire = (self.option_date - datetime.datetime.now()).total_seconds() / (
            365 * 24 * 60.0
        )

        if self.type == OptionType.CALL:
            return vollib_a.delta(
                "c",
                self.current_stock_price,
                self.strike,
                expire,
                risk,
                self.implied_volatility,
            )
        else:
            return vollib_a.delta(
                "p",
                self.current_stock_price,
                self.strike,
                expire,
                risk,
                self.implied_volatility,
            )

    def gamma(self, risk: float = 0) -> float:
        if self.type == OptionType.UNDEFINED:
            raise exceptions.OptionTypeIsUndefined()
        expire = (self.option_date - datetime.datetime.now()).total_seconds() / (
            365 * 24 * 60.0
        )

        if self.type == OptionType.CALL:
            return vollib_a.gamma(
                "c",
                self.current_stock_price,
                self.strike,
                expire,
                risk,
                self.implied_volatility,
            )
        else:
            return vollib_a.gamma(
                "p",
                self.current_stock_price,
                self.strike,
                expire,
                risk,
                self.implied_volatility,
            )

    def tetha(self, risk: float = 0) -> float:
        if self.type == OptionType.UNDEFINED:
            raise exceptions.OptionTypeIsUndefined()
        expire = (self.option_date - datetime.datetime.now()).total_seconds() / (
            365 * 24 * 60.0
        )

        if self.type == OptionType.CALL:
            return vollib_a.tetha(
                "c",
                self.current_stock_price,
                self.strike,
                expire,
                risk,
                self.implied_volatility,
            )
        else:
            return vollib_a.tetha(
                "p",
                self.current_stock_price,
                self.strike,
                expire,
                risk,
                self.implied_volatility,
            )

    def rho(self, risk: float = 0) -> float:
        if self.type == OptionType.UNDEFINED:
            raise exceptions.OptionTypeIsUndefined()
        expire = (self.option_date - datetime.datetime.now()).total_seconds() / (
            365 * 24 * 60.0
        )

        if self.type == OptionType.CALL:
            return vollib_a.rho(
                "c",
                self.current_stock_price,
                self.strike,
                expire,
                risk,
                self.implied_volatility,
            )
        else:
            return vollib_a.rho(
                "p",
                self.current_stock_price,
                self.strike,
                expire,
                risk,
                self.implied_volatility,
            )

    def vega(self, risk: float = 0) -> float:
        if self.type == OptionType.UNDEFINED:
            raise exceptions.OptionTypeIsUndefined()
        expire = (self.option_date - datetime.datetime.now()).total_seconds() / (
            365 * 24 * 60.0
        )

        if self.type == OptionType.CALL:
            return vollib_a.vega(
                "c",
                self.current_stock_price,
                self.strike,
                expire,
                risk,
                self.implied_volatility,
            )
        else:
            return vollib_a.vega(
                "p",
                self.current_stock_price,
                self.strike,
                expire,
                risk,
                self.implied_volatility,
            )

    def payoff(self, contract_size: int = 100, position: int = 1) -> float:
        # negative position means short

        if self.type == OptionType.UNDEFINED:
            raise exceptions.OptionTypeIsUndefined()

        if self.type == OptionType.CALL:
            if position > 0:
                return (
                    (max(0, self.current_stock_price - self.strike) - self.last_price)
                    * contract_size
                    * position
                )
            else:
                return (
                    (self.last_price - max(0, self.current_stock_price - self.strike))
                    * contract_size
                    * position
                )
        else:
            if position > 0:
                return (
                    (max(0, self.strike - self.current_stock_price) - self.last_price)
                    * contract_size
                    * position
                )
            else:
                return (
                    (self.last_price - max(0, self.strike - self.current_stock_price))
                    * contract_size
                    * position
                )


class OptionChain(typing.NamedTuple):
    calls: typing.List[OptionChainDatum]
    puts: typing.List[OptionChainDatum]



import typing

import pandas

from . import model, stock_data


def correlation(symbol_one: str, symbol_two: str):
    pass


def correlation_history(
    history_one: typing.List[model.StockHistoryDatum],
    history_two: typing.List[model.StockHistoryDatum],
    length=5,
) -> typing.List[model.CorrelationHistoryDatum]:

    d1 = pandas.DataFrame(data=history_one)
    d2 = pandas.DataFrame(data=history_two)

    d1 = d1.set_index("time")
    d2 = d2.set_index("time")

    merge = pandas.merge(d1, d2, on="time", suffixes=["_1", "_2"])

    df = pandas.DataFrame(merge, columns=["close_value_1", "close_value_2"])
    data = df["close_value_1"].rolling(length).corr(other=df["close_value_2"]).dropna()
    data.columns = ["value"]

    return [
        model.CorrelationHistoryDatum(time=v[1], value=v[0])
        for v in zip(data, data.index)
    ]



import datetime
import typing

import praw

from . import model, proxy


class StockData:
    symbol: str
    data_proxy: proxy.DataProxy

    def __init__(
        self, symbol: str, data_proxy: typing.Type[proxy.DataProxy], *args, **kwargs
    ):
        self.symbol = symbol
        self.data_proxy = data_proxy(symbol, *args, **kwargs)

    def get_info(self) -> model.StockInformation:
        return self.data_proxy.get_info()

    def get_last_price(self) -> float:
        return self.data_proxy.get_last_price()

    def get_stock_price_history(
        self, interval: datetime.timedelta, period: datetime.timedelta
    ) -> typing.List[model.StockHistoryDatum]:
        return self.data_proxy.get_stock_price_history(interval, period)

    def get_option_chain(self, date: datetime.datetime) -> model.OptionChain:
        return self.data_proxy.get_option_chain(date)

    def get_next_option_chain(self) -> model.OptionChain:
        return self.data_proxy.get_option_chain(date=None)

    def get_full_option_chain(
        self
    ) -> typing.Dict[datetime.datetime, model.OptionChain]:
        return self.data_proxy.get_full_option_chain()

    def get_reddit_threads(
        self, subreddits: typing.List, time_filter: str = "day", sort: str = "hot"
    ) -> typing.List[praw.models.Submission]:
        return self.data_proxy.get_reddit_threads(subreddits, time_filter, sort)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['stock_utilities']

package_data = \
{'': ['*']}

install_requires = \
['praw>=7.2.0,<8.0.0',
 'py_vollib>=1.0.1,<2.0.0',
 'scipy>=1.6.1,<2.0.0',
 'yfinance>=0.1.55,<0.2.0']

setup_kwargs = {
    'name': 'stock-utilities',
    'version': '0.1.7',
    'description': 'This library is a wrapper around the finance libraries in order to give out a data model different from raw pandas and be usable in production services',
    'long_description': '# stock_utilities\n\nThis repo will manage utilities for stock data and stock option\nThe idea is to fetch data from multiple sources and use them from a single point and have a library that is typesafe\n\n\nThe used provider are:\n  - YFinance\n\n## Example\n\n```\nimport datetime\nimport praw\nimport stock_utilities\n\ndata = stock_utilities.stock_data.StockData(\n    "GME", stock_utilities.proxy.YFinanceProvider\n)\nprint(data.get_last_price())\ndata = stock_utilities.stock_data.StockData(\n    "GME", stock_utilities.proxy.YFinanceProvider\n)\nhistory = data.get_stock_price_history(\n    interval=datetime.timedelta(days=1), period=datetime.timedelta(days=5)\n)\nassert len(history), 5\n\nhistory_option = data.get_next_friday_option_chain()\nprint(history_option.calls[-1])\nprint(\n    history_option.calls[-1].delta(),\n    history_option.calls[-1].gamma(),\n    history_option.calls[-1].vega(),\n)\n\n\ncombined_providers = stock_utilities.proxy.combine_providers(\n    [stock_utilities.proxy.YFinanceProvider, stock_utilities.proxy.RedditFetcher]\n)\n\nreddit = praw.Reddit(\n   client_id="XXX",\n   client_secret="XXX",\n   user_agent="XXX",\n)\nnew_client = stock_utilities.stock_data.StockData(\n   "GME", combined_providers, reddit_client=reddit\n)\n\nprint(new_client.get_reddit_threads(["wallstreetbets"]))\n```',
    'author': 'Diego Luca Candido',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/joxer/stock_utilities',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
